// XMLTreeCtrl.cpp : implementation file
//
//***********************************************
/* Author : Johnson Zhou(ZhouQiang)
/* Purpose: IMplementaion file for a floating tree control like
/*			list parameter infomation list .
/*I borrowed code of XML tree from Frank Le from www.codeproject.com
/*Copyright by Johnson Zhou(ZhouQiang) (c) 2002
//***********************************************/

#include "stdafx.h"
#include "vs_treectrl.h"
#include "XMLTreeCtrl.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CXMLTreeCtrl

CXMLTreeCtrl::CXMLTreeCtrl()
{
}

CXMLTreeCtrl::~CXMLTreeCtrl()
{
}


BEGIN_MESSAGE_MAP(CXMLTreeCtrl, CTreeCtrl)
	//{{AFX_MSG_MAP(CXMLTreeCtrl)
	ON_NOTIFY_REFLECT(TVN_ITEMEXPANDING, OnItemexpanding)
	ON_NOTIFY_REFLECT(NM_DBLCLK, OnDblclk)
	ON_WM_LBUTTONUP()
	ON_WM_CREATE()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CXMLTreeCtrl message handlers

BOOL CXMLTreeCtrl::PreCreateWindow(CREATESTRUCT& cs) 
{
	//TVS_HASBUTTONS|TVS_HASLINES|TVS_LINESATROOT;	
	//return CTreeCtrl::PreCreateWindow(cs);
	cs.style  |= TVS_HASBUTTONS|TVS_HASLINES|TVS_LINESATROOT;
	if ( !CTreeCtrl::PreCreateWindow (cs))
		return FALSE;
	return TRUE;
}

void CXMLTreeCtrl::OnItemexpanding(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
	HTREEITEM hItem = pNMTreeView->itemNew.hItem;
	MSXML::IXMLDOMElement* node = (MSXML::IXMLDOMElement*)/*GetTreeCtrl().*/GetItemData(hItem);
	HRESULT hr;

	*pResult = 0;
	
	// If m_bOptimizeMemory is false then we optimize by not adding and deleting 
	// already added items. The trade-off is memory exhausted!
	
	CWaitCursor waitCursor;     // This could take a while!
	/*GetTreeCtrl().*/LockWindowUpdate();
	if (pNMTreeView->action == TVE_EXPAND) {
		if (m_bOptimizeMemory == FALSE) {
			HTREEITEM hChildItem;
			if ((hChildItem = /*GetTreeCtrl().*/GetChildItem(hItem)) != NULL) {
				MSXML::IXMLDOMElement* childNode = 
					(MSXML::IXMLDOMElement*)/*GetTreeCtrl().*/GetItemData(hChildItem);
				if (childNode == NULL) {
					/*GetTreeCtrl().*/DeleteItem(hChildItem);
					MSXML::IXMLDOMNode* firstChild = NULL;
					hr = node->get_firstChild(&firstChild);
					if (SUCCEEDED(hr) && firstChild != NULL) {
						if (populateNode((MSXML::IXMLDOMElement*)firstChild, hItem) == FALSE) {
							*pResult = 1;            
						}
					}
				}
			}
		} else {
			deleteFirstChild(hItem);
			MSXML::IXMLDOMNode* firstChild = NULL;
			hr = node->get_firstChild(&firstChild);
			if (SUCCEEDED(hr) && firstChild != NULL) {
				if (populateNode((MSXML::IXMLDOMElement*)firstChild, hItem) == FALSE) {
					*pResult = 1;            
				}
			}
		}
	} else { // pNMTreeView->action == TVE_COLLAPSE
		if (m_bOptimizeMemory == TRUE) {
			deleteAllChildren(hItem);
			// Set button state.
			if (node->hasChildNodes()) {
				int nImage, nSelectedImage;
				nImage = nSelectedImage = getIconIndex(node);
				HTREEITEM hChildItem = /*GetTreeCtrl().*/InsertItem(_T(""), nImage, 
					nSelectedImage, hItem);
				/*GetTreeCtrl().*/SetItemData(hChildItem, (DWORD)NULL);
			}
		}
	}
	/*GetTreeCtrl().*/UnlockWindowUpdate();
}


BOOL CXMLTreeCtrl::loadXML(const CString &strPathName, const BOOL bOptimizeMemory /*= FALSE*/)
{
	MSXML::IXMLDOMDocument* document = NULL;
	MSXML::IXMLDOMParseError* parseError = NULL;
	MSXML::IXMLDOMElement* element = NULL;	

	DeleteAllItems();

	HRESULT hr;
	hr = CoInitialize(NULL);
	if (FAILED(hr)) {
		return FALSE;
	}

	hr = CoCreateInstance(MSXML::CLSID_DOMDocument, NULL, 
		CLSCTX_INPROC_SERVER | CLSCTX_LOCAL_SERVER, 
		MSXML::IID_IXMLDOMDocument, (LPVOID*)&document);
	if (!document) {
		return FALSE;
	}

	BSTR bstr = NULL;
	document->put_async(VARIANT_FALSE);
	bstr = strPathName.AllocSysString();
	VARIANT_BOOL varOkay = document->load(bstr);
	SysFreeString(bstr);
	
	BSTR nodeName;
	HTREEITEM hItem;	
	if (varOkay) {
		hr = document->get_documentElement(&element);
		if (FAILED(hr) || element == NULL) {
			MessageBox(_T("Empty document!"), _T("Error Loading XML"), MB_ICONWARNING);
			return FALSE;
		}
		element->get_nodeName(&nodeName);
		hItem = insertItem(element, CString(nodeName), ILI_ROOT, ILI_ROOT);	
		// Optional expand the root initially.
		Expand(hItem, TVE_EXPAND);
		Select(hItem, TVGN_CARET);
	} else {
		long line, linePos;
		BSTR reason = NULL;
		
		document->get_parseError(&parseError);
		parseError->get_errorCode(&hr);
		parseError->get_line(&line);
		parseError->get_linepos(&linePos);
		parseError->get_reason(&reason);
		
		CString strMsg;
		strMsg.Format(_T("Error 0x%.8X on line %d, position %d\r\nReason: %s"), 
			hr, line, linePos, CString(reason));
		MessageBox(strMsg, _T("Error Loading XML"), MB_ICONWARNING);
				
		SysFreeString(reason);
		return FALSE;
	}

	// Must initialize here.
	m_bOptimizeMemory = bOptimizeMemory;
	
	return TRUE;
}

// This function populates all siblings of the [in] node.
//
BOOL CXMLTreeCtrl::populateNode(MSXML::IXMLDOMElement *node, const HTREEITEM &hParent)
{
	HRESULT hr = S_OK;
	BSTR nodeType, nodeName;
	HTREEITEM hItem;

	node->get_nodeTypeString(&nodeType);
	if (!wcscmp(nodeType, L"element")) {
		node->get_nodeName(&nodeName);
		hItem = insertItem(node, CString(nodeName), ILI_ELEMENT, ILI_ELEMENT, hParent);
		populateAttributes(node, hItem);
	} else if(!wcscmp(nodeType, L"text")) {
		node->get_text(&nodeName);
		hItem = insertItem(node, CString(nodeName), ILI_TEXT, ILI_TEXT, hParent);	
	} else if(!wcscmp(nodeType, L"comment")) {
		node->get_nodeName(&nodeName);
		hItem = insertItem(node, CString(nodeName), ILI_COMMENT, ILI_COMMENT, hParent);	
	} else {    // Handle more data types here if needed.
		node->get_nodeName(&nodeName);
		hItem = insertItem(node, CString(nodeName), ILI_OTHER, ILI_OTHER, hParent);	
	}
	
	MSXML::IXMLDOMNode* nextSibling = NULL;
	hr = node->get_nextSibling(&nextSibling);
	if (SUCCEEDED(hr) && nextSibling != NULL) {
		populateNode((MSXML::IXMLDOMElement*)nextSibling, hParent);
	}

	return TRUE;
}

BOOL CXMLTreeCtrl::populateAttributes(MSXML::IXMLDOMElement *node, const HTREEITEM &hParent)
{
	HRESULT hr = S_OK;
	BSTR nodeName;
	HTREEITEM hItem;

	MSXML::IXMLDOMNamedNodeMap* namedNodeMap = NULL;
	hr = node->get_attributes(&namedNodeMap);
	if (SUCCEEDED(hr) && namedNodeMap != NULL) {
		long listLength;
		hr = namedNodeMap->get_length(&listLength);
		for(long i = 0; i < listLength; i++) {
			MSXML::IXMLDOMNode* listItem = NULL;
			hr = namedNodeMap->get_item(i, &listItem);
			listItem->get_nodeName(&nodeName);
			hItem = insertItem((MSXML::IXMLDOMElement*)listItem, 
				CString(nodeName), ILI_ATTRIBUTE, ILI_ATTRIBUTE, hParent);	
		}
	}

	return TRUE;
}

HTREEITEM CXMLTreeCtrl::insertItem(MSXML::IXMLDOMElement *node, 
	const CString &nodeName, int nImage, int nSelectedImage, 
	HTREEITEM hParent /*= TVI_ROOT*/, HTREEITEM hInsertAfter /*= TVI_LAST*/)
{
	HTREEITEM hItem = /*GetTreeCtrl().*/InsertItem(nodeName, nImage, 
		nSelectedImage, hParent, hInsertAfter);	
	/*GetTreeCtrl().*/SetItemData(hItem, (DWORD)node);

	// Set button state.
	if (node->hasChildNodes()) {
		HTREEITEM hChildItem = /*GetTreeCtrl().*/InsertItem(_T(""), nImage, 
			nSelectedImage, hItem);
		/*GetTreeCtrl().*/SetItemData(hChildItem, (DWORD)NULL);
	}

	return hItem;
}

void CXMLTreeCtrl::deleteFirstChild(const HTREEITEM &hItem)
{
	HTREEITEM hChildItem;
	if ((hChildItem = /*GetTreeCtrl().*/GetChildItem(hItem)) != NULL) {
		/*GetTreeCtrl().*/DeleteItem(hChildItem);
	}
}

void CXMLTreeCtrl::deleteAllChildren(const HTREEITEM &hItem)
{
	HTREEITEM hChildItem;
	if ((hChildItem = /*GetTreeCtrl().*/GetChildItem(hItem)) == NULL) {
		return;
	}

	do {
		HTREEITEM hNextItem = /*GetTreeCtrl().*/GetNextSiblingItem(hChildItem);
		/*GetTreeCtrl().*/DeleteItem(hChildItem);
		hChildItem = hNextItem;
	} while (hChildItem != NULL);
}

int CXMLTreeCtrl::getIconIndex(MSXML::IXMLDOMElement *node)
{
	BSTR nodeType;

	node->get_nodeTypeString(&nodeType);
	if (!wcscmp(nodeType, L"element")) {
		return ILI_ELEMENT;
	} else if(!wcscmp(nodeType, L"attribute")) {
		return ILI_ATTRIBUTE;
	} else if(!wcscmp(nodeType, L"text")) {
		return ILI_TEXT;
	} else if(!wcscmp(nodeType, L"comment")) {
		return ILI_COMMENT;
	}

	return ILI_OTHER;
}
/*
void CXMLTreeCtrl::OnItemexpanding(NMHDR* pNMHDR, LRESULT* pResult) 
{
	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
	// TODO: Add your control notification handler code here
	
	*pResult = 0;
}
*/

//DEL void CXMLTreeCtrl::OnSelchanged(NMHDR* pNMHDR, LRESULT* pResult) 
//DEL {
//DEL 	NM_TREEVIEW* pNMTreeView = (NM_TREEVIEW*)pNMHDR;
//DEL 	// TODO: Add your control notification handler code here
//DEL 	HTREEITEM sel = this->GetSelectedItem ();
//DEL 	if ( sel )
//DEL 	{
//DEL 		::ReleaseCapture ();
//DEL 		CString str = this->GetItemText (sel);
//DEL 		this->GetParent ()->SendMessage(WM_FLOAT_CTRL,(WPARAM)str.AllocSysString (),0);
//DEL 	}
//DEL 	
//DEL 	*pResult = 0;
//DEL }

void CXMLTreeCtrl::OnDblclk(NMHDR* pNMHDR, LRESULT* pResult) 
{
	CPoint pt;
	ShowMe(pt,SW_HIDE);
	*pResult = 0;
}



 void CXMLTreeCtrl::OnLButtonUp(UINT nFlags, CPoint point) 
 {
 	CRect rect;
 	GetWindowRect(&rect);
 	CPoint pt1 = point;
 	ClientToScreen(&pt1);
 	
 	if (! rect.PtInRect (pt1))
 	{
 		ShowMe(point,SW_HIDE);
 	}
 	CTreeCtrl::OnLButtonUp(nFlags, point);
 }



void CXMLTreeCtrl::SetImages(UINT resource)
{
	m_theImageList.Detach();	// For MDI
	if ( resource ==0 )
		m_theImageList.Create(IDB_BITMAP_TREE_ICONS, 16, 4, RGB(255, 255, 255));	
	else
		m_theImageList.Create (resource,16,4,RGB(255,255,255));
	SetImageList(&m_theImageList, TVSIL_NORMAL);

}

void CXMLTreeCtrl::ShowMe(CPoint &pt,int show)
{
	if ( show == SW_SHOW)
	{
		CRect rect ,parentRect,curRect;
		CPoint pt1= pt;
	//ClientToScreen(&pt1);
		rect.left = pt1.x;
		rect.top = pt1.y;
		GetWindowRect(&curRect);
		GetParent()->GetWindowRect (&parentRect);
		if( pt.x + curRect.Width ()+30 > parentRect.right )
			rect.right = pt.x - curRect.Width ();
		else
			rect.right = pt.x + curRect.Width ();

		if ( pt.y + curRect.Height ()+30 > parentRect.bottom )
			rect.bottom = pt.y -curRect.Height ();
		else
			rect.bottom = pt.y + curRect.Height ();
		rect.NormalizeRect ();
		MoveWindow(&rect);
		ShowWindow(show);
		SetFocus ();
		SetCapture ();
	}
	else
	{
		HTREEITEM sel = this->GetSelectedItem ();
 		if ( sel )
 		{
			ShowWindow (show);
			::ReleaseCapture ();
 			CString str = this->GetItemText (sel);
 			GetParent ()->SendMessage(WM_FLOAT_CTRL,(WPARAM)str.AllocSysString (),0);
  		}	
	}
}

int CXMLTreeCtrl::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CTreeCtrl::OnCreate(lpCreateStruct) == -1)
		return -1;
	
	// TODO: Add your specialized creation code here
	
	return 0;
}


BOOL CXMLTreeCtrl::CreateTree(CRect rect, CWnd *parent,UINT image)
{
	if(!CTreeCtrl::Create (WS_TABSTOP | WS_CHILD | WS_BORDER|WS_THICKFRAME   
   | TVS_HASBUTTONS | TVS_LINESATROOT | TVS_HASLINES
   | TVS_DISABLEDRAGDROP
	,rect,parent,65535))
		return FALSE;
	if (image)
		SetImages (image);
	return TRUE;
}
